/**
 * openCV_pcv_wrapper.cpp
 *
 * This is OpenCV wrapper for Point Cloud Viewer.
 * It provides two main functions need to be called in CMEX S-Function methods
 *
 * void pcvOutputs(SimStruct *S): needs to be called in mdlOutputs
 * void pcvTerminate(SimStruct *S): needs to be called in mdlTerminates
 *
 * Copyright 2012 The MathWorks, Inc.
 */

// OpenCV 2.2
#include <opencv2/opencv.hpp>
using namespace cv;

#include "sfun_pcv.h"


//---------------------------------------------------------------------------
// Macros
//---------------------------------------------------------------------------

#define VIEWER_TITLE "Point Cloud Viewer (OpenCV)"

#define ERROR_INIT_PCV "PCV Error: Failed to heap memory."



//---------------------------------------------------------------------------
// Globals
//---------------------------------------------------------------------------

static const double FOCAL_LENGTH = 580.0;

static Mat gCvImage;
static Mat gCvXyz;
static Mat gCvDepth;
static Mat gCvImageState;
static Mat gCvXyzState;
static Mat gCvDepthState;
static Mat gDestImage;
static Mat gDestDepth;
static volatile bool gIsFirstTimePerSim = true;

static char msg[1024]; // For S-Function error/warning message


//---------------------------------------------------------------------------
// Functions
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Point Cloud Viewer rendering methods using OpenCV 
// Copied from OpenCV.jp sample written by Prof. Norishige Fukushima:
// http://opencv.jp/opencv2-x-samples/point-cloud-rendering
// http://nma.web.nitech.ac.jp/fukushima/index-e.html
//---------------------------------------------------------------------------

static void xcvSetRotMatrix(CvMat* rot, double pitch, double roll, double yaw)
{
	double phi = roll / 180.0 * CV_PI;
	double theta = pitch / 180.0 * CV_PI;
	double pusai = yaw / 180.0 * CV_PI;
	
	// Calcurate shared value using trigonometric functions in advance to speed up   
	double cos_phi   = cos(phi);
	double cos_theta = cos(theta);
	double cos_pusai = cos(pusai);
	double sin_phi   = sin(phi);
	double sin_theta = sin(theta);
	double sin_pusai = sin(pusai);

	rot->data.db[0] =  cos_phi * cos_theta;
	rot->data.db[1] =  cos_phi * sin_theta * sin_pusai - sin_phi * cos_pusai;
	rot->data.db[2] =  cos_phi * sin_theta * cos_pusai + sin_phi * sin_pusai;
	rot->data.db[3] =  sin_phi * cos_theta;
	rot->data.db[4] =  sin_phi * sin_theta * sin_pusai + cos_phi * cos_pusai;
	rot->data.db[5] =  sin_phi * sin_theta * cos_pusai - cos_phi * sin_pusai;
	rot->data.db[6] = -sin_theta;
	rot->data.db[7] =  cos_theta * sin_pusai;
	rot->data.db[8] =  cos_theta * cos_pusai;
}


static void eular2rot(double pitch, double roll, double yaw, Mat& dest)
{
	if(dest.empty()) dest = Mat::zeros(3, 3, CV_64F);
	
	xcvSetRotMatrix(&CvMat(dest), pitch, roll, yaw);
}


static void lookat(Point3d from, Point3d to, Mat& destR)
{
	double x = to.x - from.x;
	double y = to.y - from.y;
	double z = to.z - from.z;

	double pitch = asin(x/sqrt(x*x+z*z)) / CV_PI * 180.0;
	double yaw   = asin(-y/sqrt(y*y+z*z)) / CV_PI * 180.0;
	
	eular2rot(pitch, 0, yaw, destR);
}


/**
 * Construct point cloud data
 * @param image
 * @param destImage
 * @param depth
 * @param destDepth
 * @param xyz
 * @param R
 * @param t
 * @param K
 * @param dist
 * @param bgColor
 * @return None
 */
template <class T>
static void projectImagefromXYZ_(Mat& image, Mat& destImage, Mat& depth, Mat& destDepth, Mat& xyz, Mat& R, Mat& t, Mat& K, Mat& dist, Scalar& bgColor)
{
	destImage.setTo(bgColor);
	destDepth.setTo(0);

	// Construct point cloud image 
    vector<Point2f> pt;
    cv::projectPoints(xyz, R, t, K, dist, pt);  
 
 	// Mapping image onto 3D points
#pragma omp parallel for
	for(int j=0; j<image.rows; j++)
	{
		int count = j * image.cols;
		for(int i=0; i<image.cols; i++, count++)
		{
    		int x = (int)(pt[count].x);
        	int y = (int)(pt[count].y);
        	if((x >= 0 && x < image.cols) && (y >= 0 && y < image.rows))
        	{
				float depthMapping = depth.at<float>(j, i);
				float depthMapped = destDepth.at<float>(y, x);
				if (depthMapped == 0 || depthMapping < depthMapped)
				{
					uchar* img = image.ptr<uchar>(j,i);
					uchar* destImg = destImage.ptr<uchar>(y,x);
            		destImg[0] = img[0];
            		destImg[1] = img[1];
            		destImg[2] = img[2];

   					destDepth.at<float>(y, x) = depthMapping;
            	}
        	}
        }
    }
}

 
/**
 * Convert Simulink data to OpenCV format data
 * @param cvImage       OpenCV Image data
 * @param cvXyz         OpenCV XYZ data
 * @param cvDepth       OpenCV Depth data
 * @param simImage      Simulink Image data
 * @param simXyz        Simulink XYZ data
 * @param frameSize     Frame size
 * @param fullFrameSize Full frame size
 * @param pixelPos      Start pixel position
 * @return None
 */ 
static void ConvSimToOpenCV(Mat& cvImage, Mat& cvXyz, Mat& cvDepth, uint8_T* simImage, real_T* simXyz, Res_T frameSize, Res_T fullFrameSize, int32_T* pixelPos)
{
	// R/G/B offset in Simulink
	int simR_Offset = 0;
	int simG_Offset = frameSize.xRes * frameSize.yRes;
	int simB_Offset = frameSize.xRes * frameSize.yRes * 2;

	// X/Y/Z offset in Simulink
	int simX_Offset = 0;
	int simY_Offset = frameSize.xRes * frameSize.yRes;
	int simZ_Offset = frameSize.xRes * frameSize.yRes * 2;
	
	int xBeginAt = pixelPos[1];
	int yBeginAt = pixelPos[0];
	int xEndAt = ((xBeginAt + frameSize.xRes) < fullFrameSize.xRes) ? (xBeginAt + frameSize.xRes) : fullFrameSize.xRes;  
	int yEndAt = ((yBeginAt + frameSize.yRes) < fullFrameSize.yRes) ? (yBeginAt + frameSize.yRes) : fullFrameSize.yRes;  

#pragma omp parallel for
	for (int y=yBeginAt; y<yEndAt; y++)
	{
		for (int x=xBeginAt; x<xEndAt; x++)
		{
			// Convert Image (RGB24 to BGR24)
			uint8_T* simImgPos = (uint8_T*)(simImage + (y - yBeginAt) + frameSize.yRes * (x - xBeginAt));
			int img_x = 3 * x;
			cvImage.at<uchar>(y,   img_x) = *(uint8_T*)(simImgPos + simB_Offset); // R
			cvImage.at<uchar>(y, ++img_x) = *(uint8_T*)(simImgPos + simG_Offset); // G
			cvImage.at<uchar>(y, ++img_x) = *(uint8_T*)(simImgPos + simR_Offset); // B

			// Convert XYZ (reverse signess of Y)
			real_T* simXyzPos = (real_T*)(simXyz + (y - yBeginAt) + frameSize.yRes * (x - xBeginAt));
			real32_T*  xyz = (real32_T*)&(cvXyz.at<float>(x + fullFrameSize.xRes * y));
			xyz[0] =  (real32_T)(*(real_T*)(simXyzPos + simX_Offset)); // X  
			xyz[1] = -(real32_T)(*(real_T*)(simXyzPos + simY_Offset)); // Y
			xyz[2] =  (real32_T)(*(real_T*)(simXyzPos + simZ_Offset)); // Z

			// Convert Depth
			cvDepth.at<float>(y, x) = xyz[2];
		}
	} 
}


/**
 * Convert RGB24 data to Simulink
 * @param  outR Red output
 * @param  outG Green output
 * @param  outB Blue output
 * @param  src  RGB24 source 
 * @return None
 */
void ConvRGB24ToSimulink(uint8_T* simR, uint8_T* simG, uint8_T* simB, Mat& cvImage)
{
    RGB24Pixel_T* img = (RGB24Pixel_T*)(cvImage.ptr<uchar>(0,0));
	
    for (int x=0; x<cvImage.cols; x++)
    {
        for (int y=0; y<cvImage.rows; y++)
        {
            RGB24Pixel_T* pBGR = (RGB24Pixel_T*)(img + x + cvImage.cols * y);
            // BGR to RGB
            *(simR++) = pBGR->nBlue;
            *(simG++) = pBGR->nGreen;
            *(simB++) = pBGR->nRed;
        }
    }
}


/**
 * Display Point Cloud Viewer. This method should be called in mdlOutputs.
 * @param S Simulink S-Function structure
 * @return None
 */
void pcvOutputs(SimStruct *S)
{
	// Static variables
	static Res_T frameSize;
	static Res_T fullFrameSize;
	static uint8_T bgColor[3];
	static Point3d viewpoint;
	static Point3d lookatpoint;
	static Mat K;
	static Mat R;
	static Mat t;
	static Mat dist;
	static int key;

	// Initialize the viewer
	if (gIsFirstTimePerSim == true)
	{
		// [y x] to [x y]
		real_T* fs = FRAME_SIZE(S);
		frameSize.xRes = (int_T)fs[1];
		frameSize.yRes = (int_T)fs[0];
		
		// [y x] to [x y]
		real_T* ffs = FULL_FRAME_SIZE(S);
		fullFrameSize.xRes = (int_T)ffs[1];
		fullFrameSize.yRes = (int_T)ffs[0];

		// Convert RGB to BGR 
		real_T* rgb = BG_COLOR(S);
		bgColor[0] = (uint8_T)rgb[2];
		bgColor[1] = (uint8_T)rgb[1];
		bgColor[2] = (uint8_T)rgb[0];
		
		real_T* lap = LOOK_AT_POINT(S);
		lookatpoint = Point3d(lap[0], lap[1], lap[2]);

    	if (SET_VIEW_POINT_BY_INPUT(S) == _ON)
    	{
    		real_T* simViewPoint = (real_T*)ssGetInputPortSignal(S, INPORT_VIEW);
    		viewpoint.x = (float)simViewPoint[0];
    		viewpoint.y = (float)simViewPoint[1];
    		viewpoint.z = (float)simViewPoint[2];
		}
		else
		{
			real_T* ivp = INITIAL_VIEW_POINT(S);
			viewpoint = Point3d(ivp[0], ivp[1], ivp[2]);
			key = 0;
		}
		
		K = Mat::eye(3, 3, CV_64F);
		if (K.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		K.at<double>(0,0) = FOCAL_LENGTH;
		K.at<double>(1,1) = FOCAL_LENGTH;
//		K.at<double>(0,2) = ((double)frameSize.xRes-1.0)*0.5;
//		K.at<double>(1,2) = ((double)frameSize.yRes-1.0)*0.5;
		K.at<double>(0,2) = ((double)fullFrameSize.xRes-1.0)*0.5;
		K.at<double>(1,2) = ((double)fullFrameSize.yRes-1.0)*0.5;
		R = Mat::eye(3, 3, CV_64F);
		if (R.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		t = Mat::zeros(3, 1, CV_64F);
		if (t.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		
		dist = Mat::zeros(Size(5,1), CV_32F);
		if (dist.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		real_T* distCoeffsParam = (real_T*)DISTORTION_COEFFISIENT(S);
		for (int i=0; i<5; i++)
		{
			dist.at<float>(0,i) = (float)distCoeffsParam[i];
		}

		// Initialize globals
		gCvImage = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes), CV_8UC3);
		if (gCvImage.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		gCvXyz = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes).area(), 1, CV_32FC3);
		if (gCvXyz.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		gCvDepth = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes), CV_32F);
		if (gCvDepth.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		gCvImageState = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes), CV_8UC3);
		if (gCvImageState.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		gCvXyzState = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes).area(), 1, CV_32FC3);
		if (gCvXyzState.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		gCvDepthState = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes), CV_32F);
		if (gCvDepthState.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		gDestImage = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes), CV_8UC3);
		if (gDestImage.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}
		gDestDepth = Mat::zeros(Size(fullFrameSize.xRes, fullFrameSize.yRes), CV_32F);
		if (gDestDepth.empty())
		{
           	ssSetErrorStatus(S, ERROR_INIT_PCV);
           	return;
		}

		// Set up viewer
		namedWindow(VIEWER_TITLE, CV_WINDOW_NORMAL);
		if (WINDOW_SIZE(S) == TRUE_SIZE)
		{
			cvResizeWindow(VIEWER_TITLE, fullFrameSize.xRes, fullFrameSize.yRes);
		}

		gIsFirstTimePerSim = false; // End of initialization
	}

	// Update viewpoint
	lookat(viewpoint, lookatpoint, R);
	t.at<double>(0,0) = viewpoint.x;
	t.at<double>(1,0) = viewpoint.y;
	t.at<double>(2,0) = viewpoint.z;
	t = R * t;

	// Project point cloud image
	boolean_T pause = *((boolean_T*)ssGetInputPortSignal(S, INPORT_PAUSE));
	if (pause)
	{
		// In case of pausing, keep using the last saved source data (states) of point cloud image
		projectImagefromXYZ_<uchar>(gCvImageState, gDestImage, gCvDepthState, gDestDepth, gCvXyzState, 
			R, t, K, dist, Scalar(bgColor[0], bgColor[1], bgColor[2]));
	}
	else
	{
		// Get XYZ and Image from Simulink for NID
    	real_T*    simXyz =  (real_T*)ssGetInputPortSignal(S, INPORT_XYZ);
    	uint8_T* simImage = (uint8_T*)ssGetInputPortSignal(S, INPORT_IMAGE);
    	
		// Convert data format from Simulink for NID to OpenCV
		int32_T* pixelPos = (int32_T*)ssGetInputPortSignal(S, INPORT_PIXEL_POS); 
    	ConvSimToOpenCV(gCvImage, gCvXyz, gCvDepth, simImage, simXyz, frameSize, fullFrameSize, pixelPos);

		projectImagefromXYZ_<uchar>(gCvImage, gDestImage, gCvDepth, gDestDepth, gCvXyz, 
			R, t, K, dist, Scalar(bgColor[0], bgColor[1], bgColor[2]));

		// Save source data for pausing
		gCvImage.copyTo(gCvImageState);
		gCvDepth.copyTo(gCvDepthState);
		gCvXyz.copyTo(gCvXyzState);
	}
	
	
	// Update the viewer 
	imshow(VIEWER_TITLE, gDestImage);
//	imshow("Image", gCvImage); // debug
//	imshow("Depth", gCvDepth); // debug

	// Output point cloud image
	if (OUTPUT_IMAGE(S) == _ON)
	{
		int npixels = fullFrameSize.xRes * fullFrameSize.yRes;
    	uint8_T* simR = (uint8_T*)ssGetOutputPortSignal(S,OUTPORT_IMAGE);
    	uint8_T* simG = (uint8_T*)(simR + npixels);
    	uint8_T* simB = (uint8_T*)(simG + npixels);
		ConvRGB24ToSimulink(simR, simG, simB, gDestImage);
	}

	// Get input for changing view point
    if (SET_VIEW_POINT_BY_INPUT(S) == _ON)
    {
		// Change view point by simulink input
    	real_T* simViewPoint = (real_T*)ssGetInputPortSignal(S, INPORT_VIEW);
    	viewpoint.x = (float)simViewPoint[0];
    	viewpoint.y = (float)simViewPoint[1];
    	viewpoint.z = (float)simViewPoint[2];
	}
    else
    {
		// Change view point by keyboard input
		real_T*	step = VIEW_POINT_STEP(S);
		if(key == 'r' || key == 'R')
		{
			real_T* ivp = INITIAL_VIEW_POINT(S);
			viewpoint.x = ivp[0];
			viewpoint.y = ivp[1];
			viewpoint.z = ivp[2];
		}
		if(key == 'k' || key == 'K')
		{
			viewpoint.y += *step;
		}
		if(key == 'j' || key == 'J' )
		{
			viewpoint.y -= *step;
		}
		if(key == 'h' || key == 'H')
		{
			viewpoint.x += *step;
		}
		if(key == 'l' || key == 'L')
		{
			viewpoint.x -= *step;
		}
		if(key == 'i' || key == 'I')
		{
			viewpoint.z += *step;
		}
		if(key == 'm' || key == 'M')
		{
			viewpoint.z -= *step;
		}
	}
	key = waitKey(1);	
}


/**
 * Terminate Point Cloud Viewer
 * @param S Simulink S-Function structure
 * @return None
 */
void pcvTerminate(SimStruct *S)
{
	// Release memory
	gCvImage.release();
	gCvXyz.release();
	gCvDepth.release();

	gCvImageState.release();
	gCvXyzState.release();
	gCvDepthState.release();

	gDestImage.release();
	gDestDepth.release();
	
	// Reset for next simulation
	gIsFirstTimePerSim = true;
}


/* [EOF] */