/**
 * sfun_pcv.cpp
 *
 * C MEX S-Function for Point Cloud Viewer.
 *
 * Copyright 2012 The MathWorks, Inc.
 */

// S-Function
#include "sfun_pcv.h"


//---------------------------------------------------------------------------
// Globals
//---------------------------------------------------------------------------

static char msg[1024]; // required for S-Funciton status message


//---------------------------------------------------------------------------
// S-function methods
//---------------------------------------------------------------------------

/**
 * Function: mdlInitializeSizes
 * Abstract:
 *    The sizes information is used by Simulink to determine the S-function
 *    block's characteristics (number of inputs, outputs, states, etc.).
 */
static void mdlInitializeSizes(SimStruct *S)
{
    // Check block parameters
    ssSetNumSFcnParams(S, NUM_OF_PARAMS);  /* Number of expected parameters */
    if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) 
    {
        ssSetErrorStatus(S, "S-Function Parameter Error: Number of parameters");
        return;
    }

	// Check Frame size parameter 
	real_T* frameSize = FRAME_SIZE(S);
	for (int i=0; i<2; i++)
	{ 
		if ((int_T)frameSize[i] < 1)
		{
        	ssSetErrorStatus(S, "S-Function Parameter Error: Frame size");
        	return;
		}
	}
	
	// Check Full frame size parameter 
	real_T* fullFrameSize = FULL_FRAME_SIZE(S);
	for (int i=0; i<2; i++)
	{ 
		if ((int_T)fullFrameSize[i] < 1 || (int_T)fullFrameSize[i] < (int_T)frameSize[i])
		{
        	ssSetErrorStatus(S, "S-Function Parameter Error: Full frame size");
        	return;
		}
	}

	// Check Background color parameter
	real_T* bgColor = BG_COLOR(S);
	for (int i=0; i<3; i++)
	{ 
		if ((int_T)bgColor[i] > 255 || (int_T)bgColor[i] < 0)
		{
        	ssSetErrorStatus(S, "S-Function Parameter Error: Background color");
		}
	}

	ssAllowSignalsWithMoreThan2D(S);
	// Configure inputs
	if (SET_VIEW_POINT_BY_INPUT(S) == _ON)
	{
		ssSetNumInputPorts(S, NUM_OF_INPUTS);
	}
	else
	{
		// No View point input
		ssSetNumInputPorts(S, NUM_OF_INPUTS-1);
	}

	// Input: Pause (ON/OFF)
	ssSetInputPortWidth(S, INPORT_PAUSE, 1);
	ssSetInputPortDataType(S, INPORT_PAUSE, SS_BOOLEAN);
   	ssSetInputPortRequiredContiguous(S, INPORT_PAUSE, true);
   	ssSetInputPortDirectFeedThrough(S, INPORT_PAUSE, 1);
   	
	// Input: Start pixel position (y x)
	ssSetInputPortWidth(S, INPORT_PIXEL_POS, 2);
	ssSetInputPortDataType(S, INPORT_PIXEL_POS, SS_INT32);
   	ssSetInputPortRequiredContiguous(S, INPORT_PIXEL_POS, true);
   	ssSetInputPortDirectFeedThrough(S, INPORT_PIXEL_POS, 1);

	// Set frame size of XYZ and Image
	DECL_AND_INIT_DIMSINFO(di);
	int_T dims[3];
	di.numDims = 3;
	dims[0] = (int_T)frameSize[0];
	dims[1] = (int_T)frameSize[1];
	dims[2] = 3;
	di.dims = dims;
	di.width = dims[0]*dims[1]*dims[2];;

	// Input: XYZ
	ssSetInputPortDimensionInfo(S, INPORT_XYZ, &di);
	ssSetInputPortDataType(S, INPORT_XYZ, SS_DOUBLE);
    ssSetInputPortRequiredContiguous(S, INPORT_XYZ, true);
    ssSetInputPortDirectFeedThrough(S, INPORT_XYZ, 1);

	// Input: Image (RGB)
	ssSetInputPortDimensionInfo(S, INPORT_IMAGE, &di);
	ssSetInputPortDataType(S, INPORT_IMAGE, SS_UINT8);
    ssSetInputPortRequiredContiguous(S, INPORT_IMAGE, true);
    ssSetInputPortDirectFeedThrough(S, INPORT_IMAGE, 1);

   	// Input: View point (x y z)
	if (SET_VIEW_POINT_BY_INPUT(S) == _ON)
	{
   		ssSetInputPortWidth(S, INPORT_VIEW, 3);
   		ssSetInputPortDataType(S, INPORT_VIEW, SS_DOUBLE);
    	ssSetInputPortRequiredContiguous(S, INPORT_VIEW, true);
    	ssSetInputPortDirectFeedThrough(S, INPORT_VIEW, 1);
    }

	// Configure outputs
	if (OUTPUT_IMAGE(S) == _ON)
	{
		ssSetNumOutputPorts(S, NUM_OF_OUTPUTS);

		// Set frame size of Image output
		DECL_AND_INIT_DIMSINFO(di);
		int_T dims[3];
		di.numDims = 3;
		dims[0] = (int_T)fullFrameSize[0];
		dims[1] = (int_T)fullFrameSize[1];
		dims[2] = 3;
		di.dims = dims;
		di.width = dims[0]*dims[1]*dims[2];
    
		// Output: Image (RGB)
		ssSetOutputPortDimensionInfo(S, OUTPORT_IMAGE, &di);
		ssSetOutputPortDataType(S, OUTPORT_IMAGE, SS_UINT8);
	}
	else
	{
		// No Outport
		ssSetNumOutputPorts(S, 0);
	}

    // Other configurations
    ssSetNumContStates(S, 0);
    ssSetNumDiscStates(S, 0);
    ssSetNumSampleTimes(S, 1);
    ssSetNumRWork(S, 0);
    ssSetNumIWork(S, 0);
    ssSetNumPWork(S, 0);
    ssSetNumModes(S, 0);
    ssSetNumNonsampledZCs(S, 0);
    ssSetSimStateCompliance(S, USE_DEFAULT_SIM_STATE);
    ssSetOptions(S, SS_OPTION_EXCEPTION_FREE_CODE);
}


/**
 * Function: mdlInitializeSampleTimes
 * Abstract:
 *    This function is used to specify the sample time(s) for your
 *    S-function. You must register the same number of sample times as
 *    specified in ssSetNumSampleTimes.
 */
static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime( S, 0, INHERITED_SAMPLE_TIME );
    ssSetOffsetTime( S, 0, 0 );
    ssSetModelReferenceSampleTimeDefaultInheritance(S);
}


/**
 * Function: mdlOutputs
 * Abstract:
 *    In this function, you compute the outputs of your S-function
 *    block.
 */
static void mdlOutputs(SimStruct *S, int_T tid)
{
	pcvOutputs(S);
}


/**
 * Function: mdlTerminate
 * Abstract:
 *    In this function, you should perform any actions that are necessary
 *    at the termination of a simulation.  For example, if memory was
 *    allocated in mdlStart, this is the place to free it.
 */
static void mdlTerminate(SimStruct *S)
{
	pcvTerminate(S);
}


//---------------------------------------------------------------------------
// Required S-function trailer
//---------------------------------------------------------------------------

#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif

/* [EOF] */
